package CompGeom;

import CompGeom.Homog;

public class Point extends Object {
  public double x, y;
  
  public Point(double px, double py) {x = px; y = py;}
  public Point(int px, int py) {x = px; y = -py;}
  public Point(Point p) {this(p.x, p.y);}
  public Point(java.awt.Point p) {this(p.x, p.y);}
  public Point(Homog p) {this(p.x/p.w, p.y/p.w);}

  static public Point center(Point a, Point b, Point c) {
    Point ctr = Point.add(b,c);
    Point ba = Point.sub(b,a);
    Point ca = Point.sub(c,a);
    double factor = ba.dot(ca)/ba.perpdot(ca);
    ctr.sub(ba.sub(ca).perp().mul(factor)).div(2.0);
    return ctr;
  }

  public boolean inCircle(Point a, Point b, Point c) {
    double ax = a.x - x, ay = a.y - y, 
           bx = b.x - x, by = b.y - y, 
           cx = c.x - x, cy = c.y - y;
    return (ax*ax + ay*ay)*(bx*cy - by*cx) 
         - (bx*bx + by*by)*(ax*cy - ay*cx) 
         + (cx*cx + cy*cy)*(ax*by - ay*bx) < 0;
  }

  public Point add(Point p) {x += p.x; y += p.y; return this;}
  public Point sub(Point p) {x -= p.x; y -= p.y; return this;}
  public Point add(java.awt.Point p) {x += p.x; y += -p.y; return this;}
  public Point sub(java.awt.Point p) {x -= p.x; y -= -p.y; return this;}
  public Point neg() {x = -x; y = -y; return this;}
  public Point mul(double m) {x *= m; y *= m; return this;}
  public Point div(double m) {x /= m; y /= m; return this;}
  public double len2() {return x*x + y*y;}
  public double len() {return Math.sqrt(this.len2());}
  final public double dist2(Point p) {return (new Point(x-p.x, y-p.y)).len2();}
  final public double dist(Point p) {return Math.sqrt(this.dist2(p));}
  public Point normalize() {return div(len());}
  static public Point add(Point p, Point q) {
    return new Point(p.x + q.x, p.y + q.y);
  }
  static public Point sub(Point p, Point q) {
    return new Point(p.x - q.x, p.y - q.y);
  }
  static public Point lincomb(Point p, double beta, Point q) {
    return new Point(p.x + beta * q.x, p.y + beta * q.y);
  }
  static public Point lincomb(double alpha, Point p, double beta, Point q) {
    return new Point(alpha * p.x + beta * q.x, alpha * p.y + beta * q.y);
  }
  final public Point perp() {double tmp = -y; y = x; x = tmp; return this;}
  public double dot(Point p) {return x*p.x + y*p.y;}
  public double  perpdot(Point p) {return x*p.y - y*p.x;}
  public double  dotperp(Point p) {return - x*p.y + y*p.x;}
  public boolean equals(Point q) {return (x == q.x) && (y == q.y);}
  public Homog meet(Point p) {
    return new Homog(x*p.y - y*p.x, y - p.y, p.x - x);
  }

 public boolean y_less_than(Point q) { /* true if p.y < q.y */
    double diff = y - q.y;
    return (diff < 0) || ((diff == 0) && (x < q.x));
  }
 public boolean less(Point q) { /* true if p.y < q.y */
    double diff = x - q.x;
    return (diff < 0) || ((diff == 0) && (y > q.y));
  }
 public boolean greater(Point q) { /* true if p.y < q.y */
    double diff = x - q.x;
    return (diff > 0) || ((diff == 0) && (y < q.y));
  }
  static public double det3(Point p, Point q, Point r) {
    return (q.x - p.x)*(r.y - p.y) - (q.y - p.y)*(r.x - p.x);
  }
  static public boolean right(Point p, Point q, Point r) {
    return det3(p, q, r) < 0;
  }
  static public boolean left(Point p, Point q, Point r) {
    return det3(p, q, r) > 0;
  }
  static public boolean colinear(Point p, Point q, Point r) {
    return det3(p, q, r) == 0;
  }
// IF r is p or q;
  static public boolean end(Point p, Point q, Point r) {
    return p.equals(r) || q.equals(r);
  }
  static public boolean right_end(Point p, Point q, Point r) {
    return end(p, q, r) || right(p, q, r);
  }
  static public boolean left_end(Point p, Point q, Point r) {
    return end(p, q, r) || left(p, q, r);
  }
// IF p,q,r colinear, return true if r on ray pq after q.
  static public boolean ahead(Point p, Point q, Point r) { 
    return (q.x - p.x)*(r.x - q.x) + (q.y - p.y)*(r.y - q.y) > 0;
  }
  static public boolean right_ahead(Point p, Point q, Point r) {
    double tmp = det3(p, q, r);
    return (tmp < 0) || ((tmp == 0) && ahead(p, q, r));
  }
  static public boolean left_ahead(Point p, Point q, Point r) {
    double tmp = det3(p, q, r);
    return (tmp > 0) || ((tmp == 0) && ahead(p, q, r));
  }
  static public boolean right_ahead_end(Point p, Point q, Point r) {
    double tmp = det3(p, q, r);
    return end(p, q, r) || (tmp < 0) || ((tmp == 0) && ahead(p, q, r));
  }
  static public boolean left_ahead_end(Point p, Point q, Point r) {
    double tmp = det3(p, q, r);
    return end(p, q, r) || (tmp > 0) || ((tmp == 0) && ahead(p, q, r));
  }
  static final int LEFT = 1;
  static final int RIGHT = 2;
  static final int END0 = 4;
  static final int END1 = 8;
  static final int ONSEG = 16;
  static final int AHEAD = 32;
  static final int BEHIND = 64;

  static public int classify(Point p, Point q, Point r) {
    double tmp = det3(p, q, r);
    if (tmp > 0) return LEFT;
    if (tmp < 0) return RIGHT;
    tmp = (q.x - p.x)*(r.x - p.x) + (q.y - p.y)*(r.y - p.y);
    if (tmp < 0) return BEHIND;
    if (tmp == 0) return END0;
    tmp = (q.x - p.x)*(r.x - q.x) + (q.y - p.y)*(r.y - q.y);
    if (tmp > 0) return AHEAD;
    if (tmp == 0) return END1;
    return ONSEG;
  }

  public String toString() { 
    return " (" + x + ", " + y + ")  ";
  }
  public java.awt.Point toJavaPoint(java.awt.Point p) { 
    p.x = (int) x; p.y = (int) -y; return p; 
  }
  public void drawFilled(java.awt.Graphics g) {
    g.fillOval((int) x-2, (int) -y-2, 4, 4); 
  }
  public void draw(java.awt.Graphics g) {
    g.drawOval((int) x-2, (int) -y-2, 4, 4);
  }
}

