package Terrain;
import CompGeom.Homog;
import java.awt.Dimension;
import java.awt.Graphics;

class duplicatePointException extends Exception {
  int edge;
  duplicatePointException(int e) { super(); this.edge = e; }
}

public class Triangulation {
  Point p[];
  int rev[];
  int free;
  public Graphics debug_gr;

  final public static int DELAUNAY = 0;
  final public static int VORONOI = 1;
  final public static int CRUST = 2;
  public static int delVorCrust = DELAUNAY;

  public Triangulation(int n) { this(-4096.0, 4096.0, -4096.0, 4096.0, n); }

  public Triangulation(double xmin, double xmax, 
		       double ymin, double ymax, int n) { 
    Point p0, p1, p2, p3;
    p0 = new Point(xmin, ymax);//  p0-p1
    p1 = new Point(xmax+1, ymax+1);//  | \ |
    p2 = new Point(xmax, ymin);//  p3-p2
    p3 = new Point(xmin, ymin);

    int maxn = 6*n;
    p = new Point[maxn]; 
    rev = new int[maxn];

    free = 0; 
    Triangle(p0,p1,p2,  6, 8, 3);
    Triangle(p0,p2,p3,  2,11,10);
    Triangle(p1,p0,p2,  0, 9, 1);
    Triangle(p2,p0,p3,  7, 5, 4);
  }
  
  public void finalize() {
  }

  public int Triangle(Point p1, Point p2, Point p3) {
    p[free++] = p1; p[free++] = p2; p[free++] = p3; 
    return free-3;
  }
  public int Triangle(Point p1, Point p2, Point p3, 
			int rev1, int rev2, int rev3) {
    rev[free] = rev1; rev[free+1] = rev2; rev[free+2] = rev3; 
    p[free++] = p1; p[free++] = p2; p[free++] = p3; 
    return free-3;
  }

  final public Point p(int i) { return p[i]; }
  final public int rev(int i) { return rev[i]; }

  final static int nextincr[] = {1,1,-2};
  final static int previncr[] = {2,-1,-1};

  final public static int next(int Tnum) { 
    return Tnum + nextincr[Tnum % 3];
  }
  final public static int prev(int Tnum) { 
    return Tnum + previncr[Tnum % 3];
  }
  final public int Vccw(int Tnum) { return rev(prev(Tnum)); }
  final public int Vcw(int Tnum) { return next(rev(Tnum)); }
  
  public int swap(int ti) {	/* Swap diagonal in triangulation */
    int tj = rev(ti), tip = prev(ti), tjp = prev(tj);
//    System.out.println("Swapping "+ti+", "+tj+" : "+tip+", "+tjp);
    p[ti] = p(tjp); p[tj] = p(tip);
    rev[rev(tjp)] = ti; rev[rev(tip)] = tj;
    rev[ti] = rev(tjp); rev[tj] = rev(tip);
    rev[tip] = tjp; rev[tjp] = tip;
    return tj;			/* returns new name of old prev(ti) */
  }

  final public boolean dead(int e) {
    Point nxt = p(next(e));
    return (rev(e) < 6) || // This is temporary to handle the boundary
      (p(e).greater(nxt) && nxt.greater(p(prev(e))));
  }

  public void traverse(Graphics g) { traverse(g, rev[0]); }
    
  public void traverse(Graphics g, int starte) { /* Traverse Tri */
    int e = starte, i = 2*free;
    int tmpe = e;
    do {
      if (dead(e) || dead(rev(e))) e = prev(e);
      else {
	drawEntEx(g, tmpe, e);
	tmpe = rev(e);
	e = prev(rev(e));
      }
    } while ((e != starte) && (i-- > 0));
  }

  static java.awt.Rectangle screen = new java.awt.Rectangle(0,0,350,350);

  public int locate(Graphics g, Point q) 
    throws duplicatePointException { return locate(g, q, rev(0)); }

  int locate(Graphics g, Point q, int hint) 
    throws duplicatePointException {
    int e = hint, i = 2*free;	/* i avoids infinite loops */
    if (!Point.left_ahead(p(e), p(next(e)), q)) {
      e = rev(e);
      if (e < 0) e = rev(0);
      System.out.println("Reversal in locate e"
			 +e+"re"+rev(e)+"h"+hint+"rh"+rev(hint));
    }
    Point p = p(e);
System.out.println("Start p:"+p+" q:"+q+" eq:"+q.equals(p));
    if (q.equals(p)) throw new duplicatePointException(e);
    do { /* termination test assumes no angle > 180 degrees */
System.out.println("Term test p:"+p(e)+" nx:"+p(next(e))+" !la:"+
(!Point.left_ahead(p(e), p(next(e)), q)));
    	if (!Point.left_ahead(p(e), p(next(e)), q)) { /* in triangle */
System.out.println("End p:"+p(next(e))+" q:"+q+" eq:"+q.equals(p(next(e))));
    		if (q.equals(p(next(e)))) throw new duplicatePointException(e);
    		else return rev(e); /* on edge or in triangle */
    		}
System.out.println("Walk ccw p:"+p(prev(e))+" !l:"+
(!Point.left(p, q, p(prev(e)))));
    	if (!Point.left(p, q, p(prev(e)))) e = Vccw(e);
    	else e = rev(next(e));
    } while (i-- > 0);
    return -1;		// We should never get here 
  }

  public void preinsert(Point p, int e1) { /* Make the two new triangles  */
    int e2 = next(e1), e3 = next(e2);      /* but leave the old untouched */
    int newTedges = free;
//    System.out.println("Preinserting "+p(e1)+p(e2)+p(e3)+rev(e1)+":"+rev(e2)+":"+rev(e3));
    Triangle(p, p(e2), p(e3), e2, rev(e2), newTedges+4);
    Triangle(p(e1), p, p(e3), e3, newTedges+2, rev(e3));
  }

  public boolean DelaunayEdge(int e) {
    return
      !p(prev(e)).inCircle(p(e), p(rev(e)), p(prev(rev(e))));
  }

  public void restoreDelaunay(int e, int last) { /* fix del in wedge */
    int i = 2*free;
    do {					 /* between e & rev(last) */
//      System.out.println("Restoring "+e+" to "+last);
      while (!DelaunayEdge(next(e))) e = swap(next(e));
      e = Vccw(e);
    } while ((e != rev(last)) && (i-- > 0));
    if (i <= 0) System.out.println("Problem in restoreDelaunay"+e);
  }

  public void insertDelaunay(int newTedges) {
    int e2, e3, e;

    e2 = rev(newTedges);	/* Finish insert of two new triangles */
    rev[e2] = newTedges;
    e = rev(newTedges+1);
    rev[e] = newTedges+1;

    e3 = next(e2);
    p[e3] = p(newTedges);
    rev[e3] = newTedges+3;
    e = rev(newTedges+5);
    rev[e] = newTedges+5;

    restoreDelaunay(e3, rev(e3));
  }

  public void insert(Graphics g, Point q) {
    int newTedges = free;
    try{
      preinsert(q, locate(g, q));
      insertDelaunay(newTedges);
    }
    catch(duplicatePointException e) {
      System.out.println("Point "+q+" already present"); 
    }
  }

  public void deleteTriangle(int e) { p[3*(e/3)] = null; }

  public void delete(Graphics g, Point q) {
    int e, first, last;
    int i = 2*free, j;
    try { e = locate(g, q); } 
    catch(duplicatePointException except) { e = except.edge; }
    if (q.dist2(p(e)) > q.dist2(p(next(e)))) e = prev(e); /* e not closest */
    if (q.dist2(p(e)) > q.dist2(p(prev(e)))) e = prev(e);
//    System.out.println("Delete "+q+e+p(e));

    if (q.equals(p(0)) || q.equals(p(1)) || q.equals(p(2)) || q.equals(p(5))) {
      System.out.println("Can't delete BBox corners");
      return;
    }

    while ((e != Vccw(Vccw(e))) && (i-- > 0)) {
      j = 2*free;
      while ((!Point.right(p(prev(e)), p(next(e)), p(prev(rev(e))))) 
	     && (j-- > 0)) e = Vccw(e);
//      System.out.println("Found convex "+e+p(e)+next(e)+" "+prev(rev(e)));
      if (j <= 0) System.out.println("Problem in delete (inner)"+e+p(e)+rev(e));
      e = next(swap(e));
      rd(g);
      last = next(e); first = Vcw(last); restoreDelaunay(first, rev(last));
      first = rev(last); last = Vccw(first); restoreDelaunay(first, rev(last));
      rd(g);
    }

    if (i <= 0) System.out.println("Problem in delete (outer)"+e+p(e)+rev(e));
    last = Vccw(e);
    int rne = rev(next(e));
    int rnlast = rev(next(last));
//    System.out.println("Done "+e+"/"+Vccw(e)+ " keeping "+rne+"/"+rnlast);
    if (rne >= 0) rev[rne] = rnlast;
    if (rnlast >= 0) rev[rnlast] = rne;
    deleteTriangle(rev(e));
    deleteTriangle(e);
    rd(g);
    if (rne >= 0) { e = prev(rne); restoreDelaunay(e, rev(Vccw(e))); }
    rd(g);
  }

  void rd(Graphics g) {
      g.setColor(java.awt.Color.white);
      g.fillRect(0,0,450,450);
      g.setColor(java.awt.Color.black);
      draw(g);
      try {Thread.sleep(300); } catch (InterruptedException exc) {}
    }

  public void draw(Graphics g, int e1, int e2) {
    g.drawLine((int) p(e1).x, (int) -p(e1).y, (int) p(e2).x, (int) -p(e2).y); 
  }

  public void drawEntEx(Graphics g, int ent, int ex) {
    g.setColor(java.awt.Color.red);
    g.drawLine((int)  ((2*p(ent).x+p(next(ent)).x)/3.0), 
	       (int) -((2*p(ent).y+p(next(ent)).y)/3.0), 
	       (int)  ((2*p(next(ex)).x+p(ex).x)/3.0), 
	       (int) -((2*p(next(ex)).y+p(ex).y)/3.0)); 
    g.setColor(java.awt.Color.black);
  }

  public void draw(Graphics g) {
    for (int i = 6; i < free; i++) 
      if (p(i) == null) i += 2;
      else {
	draw(g, i);
/*	if (i%3 == 0) g.drawString("f"+(i/3), 
				 (int) ((p(i).x+p(i+1).x+p(i+2).x)/3),
				-(int) ((p(i).y+p(i+1).y+p(i+2).y)/3));/**/
      }
    g.setColor(java.awt.Color.black);
    for (int i = 6; i < free; i++) p(i).draw(g);
  }

  final public void drawOld(Graphics g, int e) {
    int nxt = next(e);
    if (true) {
      if (Math.abs(p(e).x - p(nxt).x) > Math.abs(p(e).y - p(nxt).y)) {
	g.drawLine((int) p(e).x, (int) -p(e).y+2, 
		   (int) p(nxt).x, (int) -p(nxt).y+2); 
	g.drawLine((int) p(e).x, (int) -p(e).y-2, 
		   (int) p(nxt).x, (int) -p(nxt).y-2); 
      }
      else {
	g.drawLine((int) p(e).x+2, (int) -p(e).y, 
		   (int) p(nxt).x+2, (int) -p(nxt).y); 
	g.drawLine((int) p(e).x-2, (int) -p(e).y, 
		   (int) p(nxt).x-2, (int) -p(nxt).y); 
      }
    }
/*      if (!DelaunayEdge(e)) System.out.println("Non D edge "+e);
 if ((rev(e) >= 0) && (rev(rev(e)) != e)) 
 System.out.println("Bad rev2("+e+")=rev("+rev(e)+")="+rev(rev(e)));/* */
      g.drawLine((int) p(e).x, (int) -p(e).y,
		 (int) p(nxt).x, (int) -p(nxt).y); 
  }
 

  final public void draw(Graphics g, CompGeom.Point p, CompGeom.Point q) {
    g.drawLine((int) p.x, (int) -p.y, (int) q.x, (int) -q.y); 
  }


  final public void draw(Graphics g, int e) { /* draw crust or anti-crust */
    boolean drawDelaunay = false;
    int nxt = next(e); int le = prev(e); int re = prev(rev(e));
    if ((e < 6) || (delVorCrust == DELAUNAY)) drawDelaunay = true;
    else if (delVorCrust == CRUST) {
      CompGeom.Point pq = CompGeom.Point.sub(p(le), p(e));
      CompGeom.Point sr = CompGeom.Point.sub(p(re), p(nxt));
      CompGeom.Point rq = CompGeom.Point.sub(p(nxt), p(e)).perp();
      drawDelaunay = ((pq.dot(p(le))-pq.dot(p(nxt)))
		      *(sr.dot(p(re))-sr.dot(p(e))) 	
		      >= -pq.dot(rq)*sr.dot(rq)); 
    }
    if (drawDelaunay) {		/* Draw Delaunay */
	g.setColor(java.awt.Color.blue);
	draw(g, e, nxt);
      } else {			/* Draw Voronoi edge */
	g.setColor(java.awt.Color.red);
	draw(g, CompGeom.Point.center(p(le),p(e),p(nxt)),
	     CompGeom.Point.center(p(re),p(nxt),p(e)));
      }
  }


}

